Attribute VB_Name = "Module1"
Option Explicit

Public s() As Single
Public Density() As Single
    
Public Vx() As Single
Public Vy() As Single
Public Vz() As Single
    
Public Vx0() As Single
Public Vy0() As Single
Public Vz0() As Single

Public N As Long
Public TimeStep As Single
Public Diff As Single
Public Visc As Single

Public Function Offset(X As Long, Y As Long, Z As Long)

Offset = (X + (Y * N) + (Z * N * N))

End Function

Public Sub Init(Size As Long, Diffusion As Long, Viscosity As Long, Time As Single)

Dim NCube As Long

N = Size
TimeStep = Time
Diff = Diffusion
Visc = Viscosity

NCube = N ^ 3

ReDim s(NCube)
ReDim Density(NCube)

ReDim Vx(NCube)
ReDim Vy(NCube)
ReDim Vz(NCube)

ReDim Vx0(NCube)
ReDim Vy0(NCube)
ReDim Vz0(NCube)

End Sub

Public Sub SetBound(b As Long, ByRef X() As Single)

Dim i As Long
Dim j As Long
Dim k As Long

For i = 1 To N - 2
    For j = 1 To N - 2
        X(Offset(i, j, 0)) = IIf(b = 3, -X(Offset(i, j, 1)), X(Offset(i, j, 1)))
        X(Offset(i, j, N - 1)) = IIf(b = 3, -X(Offset(i, j, N - 2)), X(Offset(i, j, N - 2)))
    Next
Next

For k = 1 To N - 2
    For i = 1 To N - 2
        X(Offset(i, 0, k)) = IIf(b = 2, -X(Offset(i, 1, k)), X(Offset(i, 1, k)))
        X(Offset(i, N - 1, k)) = IIf(b = 2, -X(Offset(i, N - 2, k)), X(Offset(i, N - 2, k)))
    Next
Next

For k = 1 To N - 2
    For j = 1 To N - 2
        X(Offset(0, j, k)) = IIf(b = 1, -X(Offset(1, j, k)), X(Offset(1, j, k)))
        X(Offset(N - 1, j, k)) = IIf(b = 1, -X(Offset(N - 2, j, k)), X(Offset(N - 2, j, k)))
    Next
Next

X(Offset(0, 0, 0)) = 0.33 * (X(Offset(1, 0, 0)) + X(Offset(0, 1, 0)) + X(Offset(0, 0, 1)))
X(Offset(0, N - 1, 0)) = 0.33 * (X(Offset(1, N - 1, 0)) + X(Offset(0, N - 2, 0)) + X(Offset(0, N - 1, 1)))
X(Offset(0, 0, N - 1)) = 0.33 * (X(Offset(1, 0, N - 1)) + X(Offset(0, 1, N - 1)) + X(Offset(0, 0, N)))
X(Offset(0, N - 1, N - 1)) = 0.33 * (X(Offset(1, N - 1, N - 1)) + X(Offset(0, N - 2, N - 1)) + X(Offset(0, N - 1, N - 2)))
X(Offset(N - 1, 0, 0)) = 0.33 * (X(Offset(N - 2, 0, 0)) + X(Offset(N - 1, 1, 0)) + X(Offset(N - 1, 0, 1)))
X(Offset(N - 1, N - 1, 0)) = 0.33 * (X(Offset(N - 2, N - 1, 0)) + X(Offset(N - 1, N - 2, 0)) + X(Offset(N - 1, N - 1, 1)))
X(Offset(N - 1, 0, N - 1)) = 0.33 * (X(Offset(N - 2, 0, N - 1)) + X(Offset(N - 1, 1, N - 1)) + X(Offset(N - 1, 0, N - 2)))
X(Offset(N - 1, N - 1, N - 1)) = 0.33 * (X(Offset(N - 2, N - 1, N - 1)) + X(Offset(N - 1, N - 2, N - 1)) + X(Offset(N - 1, N - 1, N - 2)))

End Sub

Public Sub LinSolve(b As Long, ByRef X() As Single, ByRef x0() As Single, a As Single, c As Single, Iter As Long)

Dim cRecip As Single
Dim i As Long
Dim j As Long
Dim k As Long
Dim m As Long

cRecip = 1 / c

For k = 0 To Iter - 1
    For m = 1 To N - 2
        For j = 1 To N - 2
            For i = 1 To N - 2
                X(Offset(i, j, m)) = (x0(Offset(i, j, m)) + a * (X(Offset(i + 1, j, m)) + X(Offset(i - 1, j, m)) + X(Offset(i, j + 1, m)) + X(Offset(i, j - 1, m)) + X(Offset(i, j, m + 1)) + X(Offset(i, j, m - 1)))) * cRecip
            Next
        Next
    Next
    Call SetBound(b, X)
Next

End Sub

Public Sub Diffuse(b As Long, ByRef X() As Single, ByRef x0() As Single, Diffu As Single, dt As Single, Iter As Long)

Dim a As Single

a = dt * Diffu * (N - 2) * (N - 2)
Call LinSolve(b, X, x0, a, 1 + (6 * a), Iter)

End Sub

Public Sub Advect(b As Long, ByRef d() As Single, ByRef d0() As Single, ByRef VelocX() As Single, ByRef VelocY() As Single, ByRef VelocZ() As Single, dt As Single)

Dim i0 As Single
Dim i1 As Single
Dim j0 As Single
Dim j1 As Single
Dim k0 As Single
Dim k1 As Single

Dim s0 As Single
Dim s1 As Single
Dim t0 As Single
Dim t1 As Single
Dim u0 As Single
Dim u1 As Single

Dim Tmp1 As Single
Dim Tmp2 As Single
Dim Tmp3 As Single
Dim X As Single
Dim Y As Single
Dim Z As Single

Dim NFloat As Single

Dim iFloat As Single
Dim jFloat As Single
Dim kFloat As Single

Dim i As Long
Dim j As Long
Dim k As Long

Dim dtX As Single
Dim dtY As Single
Dim dtZ As Single

Dim i0i As Long
Dim i1i As Long
Dim j0i As Long
Dim j1i As Long
Dim k0i As Long
Dim k1i As Long

NFloat = N

dtX = dt * (N - 2)
dtY = dt * (N - 2)
dtZ = dt * (N - 2)

kFloat = 0
For k = 1 To N - 2
    kFloat = kFloat + 1
    
    jFloat = 0
    For j = 1 To N - 2
        jFloat = jFloat + 1
        
        iFloat = 0
        For i = 1 To N - 2
            iFloat = iFloat + 1
            
            Tmp1 = dtX * VelocX(Offset(i, j, k))
            Tmp2 = dtY * VelocY(Offset(i, j, k))
            Tmp3 = dtZ * VelocZ(Offset(i, j, k))
            
            X = iFloat - Tmp1
            Y = jFloat - Tmp2
            Z = kFloat - Tmp3
            
            If X < 0.5 Then X = 0.5
            If X > NFloat + 0.5 Then X = NFloat + 0.5
            i0 = Fix(X)
            i1 = i0 + 1
            If Y < 0.5 Then Y = 0.5
            If Y > NFloat + 0.5 Then Y = NFloat + 0.5
            j0 = Fix(Y)
            j1 = j0 + 1
            If Z < 0.5 Then Z = 0.5
            If Z > NFloat + 0.5 Then Z = NFloat + 0.5
            k0 = Fix(Z)
            k1 = k0 + 1
            
            s1 = X - i0
            s0 = 1 - s1
            t1 = Y - j0
            t0 = 1 - t1
            u1 = Z - k0
            u0 = 1 - u1
            
            i0i = i0
            i1i = i1
            j0i = j0
            j1i = j1
            k0i = k0
            k1i = k1
            
            d(Offset(i, j, k)) = s0 * (t0 * (u0 * d0(Offset(i0i, j0i, k0i)) _
                                           + u1 * d0(Offset(i0i, j0i, k1i))) _
                                    + (t1 * (u0 * d0(Offset(i0i, j1i, k0i)) _
                                           + u1 * d0(Offset(i0i, j1i, k1i))))) _
                               + s1 * (t0 * (u0 * d0(Offset(i1i, j0i, k0i)) _
                                           + u1 * d0(Offset(i1i, j0i, k1i))) _
                                    + (t1 * (u0 * d0(Offset(i1i, j1i, k0i)) _
                                           + u1 * d0(Offset(i1i, j1i, k1i)))))
        Next
    Next
Next

Call SetBound(b, d)

End Sub

Public Sub Project(ByRef VelocX() As Single, ByRef VelocY() As Single, ByRef VelocZ() As Single, ByRef p() As Single, ByRef Div() As Single, Iter As Long)

Dim i As Long
Dim j As Long
Dim k As Long

For k = 1 To N - 2
    For j = 1 To N - 2
        For i = 1 To N - 2
            Div(Offset(i, j, k)) = -0.5 * (VelocX(Offset(i + 1, j, k)) - VelocX(Offset(i - 1, j, k)) + VelocY(Offset(i, j + 1, k)) - VelocY(Offset(i, j - 1, k)) + VelocZ(Offset(i, j, k + 1)) - VelocZ(Offset(i, j, k - 1))) / N
            p(Offset(i, j, k)) = 0
        Next
    Next
Next

Call SetBound(0, Div)
Call SetBound(0, p)
Call LinSolve(0, p, Div, 1, 6, Iter)

For k = 1 To N - 2
    For j = 1 To N - 2
        For i = 1 To N - 2
            VelocX(Offset(i, j, k)) = VelocX(Offset(i, j, k)) - (0.5 * (p(Offset(i + 1, j, k)) - p(Offset(i - 1, j, k))) * N)
            VelocY(Offset(i, j, k)) = VelocY(Offset(i, j, k)) - (0.5 * (p(Offset(i, j + 1, k)) - p(Offset(i, j - 1, k))) * N)
            VelocZ(Offset(i, j, k)) = VelocZ(Offset(i, j, k)) - (0.5 * (p(Offset(i, j, k + 1)) - p(Offset(i, j, k - 1))) * N)
        Next
    Next
Next

Call SetBound(1, VelocX)
Call SetBound(2, VelocY)
Call SetBound(3, VelocZ)

End Sub

Public Sub FluidCubeStep()

Dim X As Long
Dim Y As Long
Dim BaseColor As Long

Call Diffuse(1, Vx0, Vx, Visc, TimeStep, 4)
Call Diffuse(2, Vy0, Vy, Visc, TimeStep, 4)
Call Diffuse(3, Vz0, Vz, Visc, TimeStep, 4)

Call Project(Vx0, Vy0, Vz0, Vx, Vy, 4)

Call Advect(1, Vx, Vx0, Vx0, Vy0, Vz0, TimeStep)
Call Advect(2, Vy, Vy0, Vx0, Vy0, Vz0, TimeStep)
Call Advect(3, Vz, Vz0, Vx0, Vy0, Vz0, TimeStep)

Call Project(Vx, Vy, Vz, Vx0, Vy0, 4)

Call Diffuse(0, s, Density, Diff, TimeStep, 4)
Call Advect(0, Density, s, Vx, Vy, Vz, TimeStep)

'Do drawing here

For X = 1 To N - 2
    For Y = 1 To N - 2
        BaseColor = Density(Offset(X, Y, N / 2)) * 1000000
        If BaseColor > 255 Then BaseColor = 255
        Form1.Picture1.Line (X, Y)-(X + 1, Y + 1), RGB(BaseColor, 255 - BaseColor, 255), BF
    Next
Next

DoEvents

End Sub

Public Sub FluidCubeAddDensity(X As Long, Y As Long, Z As Long, Amount As Single)

Density(Offset(X, Y, Z)) = Density(Offset(X, Y, Z)) + Amount

End Sub

Public Sub FluidCubeAddVelocity(X As Long, Y As Long, Z As Long, AmountX As Single, AmountY As Single, AmountZ As Single)

Dim Index As Long

Index = Offset(X, Y, Z)

Vx(Index) = Vx(Index) + AmountX
Vy(Index) = Vy(Index) + AmountY
Vz(Index) = Vz(Index) + AmountZ

End Sub













